/*
 * Copyright (c) 2001-2004 Swedish Institute of Computer Science.
 * All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE.
 *
 * This file is part of the lwIP TCP/IP stack.
 * 
 * Author: Sandeep Dutta
 *
 */

/*
 * This file is a Ethernet network interface drivers for lwIP using the EASY MAC
 */
#include <manik_system.h>
#include "lwip/opt.h"
#include "lwip/def.h"
#include "lwip/mem.h"
#include "lwip/pbuf.h"
#include "lwip/sys.h"
#include <lwip/stats.h>

#include "netif/etharp.h"
#include "easymacif.h"

#define eth_base (volatile unsigned int *)(EEMAC_BASE)
#define __MIN_ETH_PACKET__ 60

/* Define those to better describe your network interface. */
#define IFNAME0 'e'
#define IFNAME1 'n'

struct easymacif {
  struct eth_addr *ethaddr;
  /* Add whatever per-interface state that is needed here. */
};

static const struct eth_addr ethbroadcast = {{0xff,0xff,0xff,0xff,0xff,0xff}};

/* Forward declarations. */
static void  easymacif_input(struct netif *netif);
static err_t easymacif_output(struct netif *netif, struct pbuf *p,
			      struct ip_addr *ipaddr);

/*--------------------------------------------------------------*/
/* eth_get_status() - returns the status . the bits are     	*/
/* 		      BIT:0 - 0					*/
/* 		      BIT:1 - receive enabled			*/
/* 		      BIT:2 - transmit busy 			*/
/* 		      BIT:3 - 0		 			*/
/* 		      BIT:4 - 0		 			*/
/* 		      BIT:5 - receive interrupt enabled		*/
/* 		      BIT:6 - transmit interrupt enabled	*/
/* 		      BIT:7 - promiscuous mode enabled		*/
/* 		      BIT:8 - packet ready in receive buffer	*/
/*		      BIT:9 - CRC error valid only when BIT:8   */
/* 		      BIT:10-15 - 0  				*/
/*		      BIT:31-16 - receive packet length valid when BIT:8 */
/*--------------------------------------------------------------*/
static unsigned int eth_get_status()
{
	unsigned int estat = *(eth_base+EEMAC_CTRL_REG);
	return estat;
}

/*--------------------------------------------------------------*/
/* eth_set_status() - set the ethernet status bits              */
/*--------------------------------------------------------------*/
inline static  void eth_set_status(unsigned int stat_bits)
{
	*(eth_base+EEMAC_CTRL_REG) |= stat_bits;
}

/*--------------------------------------------------------------*/
/* eth_set_macaddr() - sets the mac address 			*/
/*--------------------------------------------------------------*/
static void eth_set_macaddr( char *macaddr, int len)
{
	int i;
	
	eth_set_status(EEMAC_RESET_MPTR);
	
	for (i = 0; i < len; i++)
		*(eth_base+EEMAC_MACADDR_REG) = *macaddr--;	
}

/*--------------------------------------------------------------*/
/* eth_clr_status() - clear bits in the status register         */
/*	    returns - previous valus of status bits		*/
/*--------------------------------------------------------------*/
static unsigned int eth_clr_status(unsigned int stat_bits)
{
	unsigned int prev;

	prev = *(eth_base+EEMAC_CTRL_REG);
	*(eth_base+EEMAC_CTRL_REG) = (prev & ~stat_bits);
	return prev;
}

/*--------------------------------------------------------------*/
/* eth_get_len() - returns length of received packet		*/
/*--------------------------------------------------------------*/
static int eth_get_len()
{
	return eth_get_status() >> 16;
}

static void eth_print_stat()
{
	unsigned int estat = eth_get_status();
	short elen = estat >> 16;
	
	estat &= 0x0000ffff;
	if (estat & EEMAC_RECV_ENB)    puts("Receive enabled\n\r");
	if (estat & EEMAC_XMIT_BUSY)   puts("Transmit in progress\n\r");
	if (estat & EEMAC_RECV_IENB)   puts("Receive interrupt enabled\n\r");
	if (estat & EEMAC_XMIT_IENB)   puts("Transmit interrupt enabled\n\r");
	if ((estat & (EEMAC_PROMISCUOUS|EEMAC_RECV_DONE)) == (EEMAC_PROMISCUOUS|EEMAC_RECV_DONE))
		puts("Promiscuous mode enabled\n\r");
	if (estat & EEMAC_RECV_DONE)  puts("Packet received\n\r");
	if (estat & EEMAC_RECV_CRCERR) puts("CRC Error detected\n\r");
}

/*--------------------------------------------------------------*/
/* eth_isr() - interrupt service routine called by crt0         */
/*--------------------------------------------------------------*/
void easymacif_isr(int *registers)
{
	eth_clr_status(EEMAC_RECV_IENB); /* disable interrupts */
}

/*--------------------------------------------------------------*/
/* low_level_init :						*/
/*--------------------------------------------------------------*/
static void low_level_init(struct netif *netif)
{
	int i = 0 ;
	unsigned int stat ;

	/* set MAC hardware address length */
	netif->hwaddr_len = 6;

	/* set MAC hardware address */
	for (i = 0 ; i < 6; i++ )
		netif->hwaddr[i] = i;

	/* maximum transfer unit */
	netif->mtu = 1500;
	
	/* broadcast capability */
	netif->flags = NETIF_FLAG_BROADCAST;
 
	/* set mac address on the interface */
	eth_set_macaddr((char *)netif->hwaddr[5],6);
	
	/* register isr */
	register_isr(EXTRN_IRQ,easymacif_isr,EEMAC_IRQ);

	/* enable promiscuous mode, recv & reset mac address pointer & 
	   enable receive interrupt */ 
	eth_set_status(EEMAC_RECV_ENB|EEMAC_PROMISCUOUS|EEMAC_RESET_COUNTER|EEMAC_RECV_IENB);
	
	/* enable in the PSW */
	if (EEMAC_IRQ == 0) stat = EI0_ENB;
	else if (EEMAC_IRQ == 1) stat = EI1_ENB;
	else if (EEMAC_IRQ == 2) stat = EI2_ENB;
	else if (EEMAC_IRQ == 3) stat = EI3_ENB;
	else if (EEMAC_IRQ == 4) stat = EI4_ENB;
	else if (EEMAC_IRQ == 5) stat = EI5_ENB;
	SET_PSWBIT(IE_FLAG|stat);
}

/*
 * low_level_output():
 *
 * Does the actual transmission of the packet. The packet is
 * contained in the pbuf that is passed to the function. This pbuf
 * might be chained.
 *
 */

static err_t low_level_output(struct netif *netif, struct pbuf *p)
{
	struct easymacif *ethernetif = (struct easymacif *) netif->state;
	struct pbuf *q;
	int total_len = 0;

	/* wait for xmit buffer to become empty */
	while (eth_get_status() & EEMAC_XMIT_BUSY);
  
#if ETH_PAD_SIZE
	pbuf_header(p, -ETH_PAD_SIZE);			/* drop the padding word */
#endif
	/* reset the write pointer */
	eth_set_status(EEMAC_RESET_WPTR);
	for(q = p; q != NULL; q = q->next) {
		int i ;
		/* Send the data from the pbuf to the interface, one pbuf at a
		   time. The size of the data in each pbuf is kept in the ->len
		   variable. */
		for (i = 0 ; i < q->len; i++) {
			*(eth_base+EEMAC_DATA_REG) = ((char *)q->payload)[i];
		}
		total_len += q->len;
	}
	/* pad packets < minimum size with zeros */
	for (;total_len <= __MIN_ETH_PACKET__; total_len++) {
		*(eth_base+EEMAC_DATA_REG) = '\0';
	}
	/* start transmission */
	eth_set_status(EEMAC_START_XMIT);

#if ETH_PAD_SIZE
	pbuf_header(p, ETH_PAD_SIZE);			/* reclaim the padding word */
#endif
  
#if LINK_STATS
	lwip_stats.link.xmit++;
#endif /* LINK_STATS */      
	
	return ERR_OK;
}

/*
 * low_level_input():
 *
 * Should allocate a pbuf and transfer the bytes of the incoming
 * packet from the interface into the pbuf.
 *
 */

static struct pbuf * low_level_input(struct netif *netif)
{
	struct easymacif *ethernetif = netif->state;
	struct pbuf *p, *q;
	u16_t len;

	/* Obtain the size of the packet and put it into the "len" variable. */
	len = eth_get_len();

#if ETH_PAD_SIZE
	len += ETH_PAD_SIZE;	  /* allow room for Ethernet padding */
#endif

	/* We allocate a pbuf chain of pbufs from the pool. */
	p = pbuf_alloc(PBUF_RAW, len, PBUF_POOL);
  
	if (p != NULL) {

#if ETH_PAD_SIZE
		pbuf_header(p, -ETH_PAD_SIZE);			/* drop the padding word */
#endif
		
		/* We iterate over the pbuf chain until we have read the entire
		 * packet into the pbuf. */
		for(q = p; q != NULL; q = q->next) {
			int i;
			/* Read enough bytes to fill this pbuf in the chain. The
			 * available data in the pbuf is given by the q->len
			 * variable. */
			for (i = 0 ; i < q->len ; i++)
				((char *)q->payload)[i] = *(eth_base+EEMAC_DATA_REG);
		}
		/* read complete & re-enable interrupts */
		eth_set_status(EEMAC_READ_DONE|EEMAC_RECV_IENB);

#if ETH_PAD_SIZE
		pbuf_header(p, ETH_PAD_SIZE);			/* reclaim the padding word */
#endif

#if LINK_STATS
		lwip_stats.link.recv++;
#endif /* LINK_STATS */      
	} else {
		/* read complete & re-enable interrupts */
		eth_set_status(EEMAC_READ_DONE|EEMAC_RECV_IENB);
#if LINK_STATS
		lwip_stats.link.memerr++;
		lwip_stats.link.drop++;
#endif /* LINK_STATS */      
	}
	
	return p;  
}

/*
 * easymacif_output():
 *
 * This function is called by the TCP/IP stack when an IP packet
 * should be sent. It calls the function called low_level_output() to
 * do the actual transmission of the packet.
 *
 */
static err_t easymacif_output(struct netif *netif, struct pbuf *p,
			      struct ip_addr *ipaddr)
{	
	/* resolve hardware address, then send (or queue) packet */
	return etharp_output(netif, ipaddr, p);
}

/*
 * easymacif_input():
 *
 * This function should be called when a packet is ready to be read
 * from the interface. It uses the function low_level_input() that
 * should handle the actual reception of bytes from the network
 * interface.
 *
 */

static void easymacif_input(struct netif *netif)
{
	struct easymacif *ethernetif;
	struct eth_hdr *ethhdr;
	struct pbuf *p;
	
	ethernetif = netif->state;
  
	/* move received packet into a new pbuf */
	p = low_level_input(netif);
	/* no packet could be read, silently ignore this */
	if (p == NULL) return;
	/* points to packet payload, which starts with an Ethernet header */
	ethhdr = p->payload;

#if LINK_STATS
	lwip_stats.link.recv++;
#endif /* LINK_STATS */

	ethhdr = p->payload;
    
	switch (htons(ethhdr->type)) {
		/* IP packet? */
	case ETHTYPE_IP:
		/* skip Ethernet header */
		pbuf_header(p, -sizeof(struct eth_hdr));
		/* pass to network layer */
		netif->input(p, netif);
		break;
      
	case ETHTYPE_ARP:
		/* pass p to ARP module  */
		etharp_arp_input(netif, ethernetif->ethaddr, p);
		break;
	default:
		pbuf_free(p);
		p = NULL;
		break;
	}
}

unsigned char junk[] = {0x0, 0x50, 0x5b, 0x2, 0x97, 0xcb, 0x0, 0x1, 0x2, 0x3, 0x4, 0x5, 
			0x8, 0x00, 0x45, 0x0, 0x00, 0x2c, 0x0, 0x0, 0x40, 0x0, 0xff, 0x6, 
			0xf5, 0xbf, 0xc0, 0xa8, 0x2, 0x63, 0xc0, 0xa8, 0x2, 0x58, 0x0, 0x50, 
			0x7, 0x25, 0x0, 0x0, 0x19, 0x73, 0xb2, 0x33, 0xd4, 0x94, 0x60, 0x12, 
			0x8, 0x0, 0x67, 0x8e, 0x0, 0x0, 0x2, 0x4, 0x0, 0x80 };

/*--------------------------------------------------------------*/
/* eth_send_packet() - sends a packet, waits for xmit buffer to */
/*                     to become empty				*/
/*--------------------------------------------------------------*/
static void eth_send_packet(unsigned char *buff, int len)
{
	int i;
	
	printf("len = %d\n",len);
	/* wait for xmit buffer to become empty */
	while (eth_get_status() & EEMAC_XMIT_BUSY);
	
	/* reset the write pointer */
	*(eth_base+EEMAC_CTRL_REG) |= EEMAC_RESET_WPTR;
	for (i = 0 ; i < len; i++ )
		*(eth_base+EEMAC_DATA_REG) = *buff++;

	/* start transmission */
	*(eth_base+EEMAC_CTRL_REG) |= EEMAC_START_XMIT;	
}

/*
 * easymacif_init():
 *
 * Should be called at the beginning of the program to set up the
 * network interface. It calls the function low_level_init() to do the
 * actual setup of the hardware.
 *
 */

err_t easymacif_init(struct netif *netif)
{
	struct easymacif *easymacif;
    
	easymacif = mem_malloc(sizeof(struct easymacif));
  
	if (easymacif == NULL) {
		LWIP_DEBUGF(NETIF_DEBUG, ("ethernetif_init: out of memory\n"));
		return ERR_MEM;
	}
  
	netif->state = easymacif;
	netif->name[0] = IFNAME0;
	netif->name[1] = IFNAME1;
	netif->output = easymacif_output;
	netif->linkoutput = low_level_output;

	easymacif->ethaddr = (struct eth_addr *)&(netif->hwaddr[0]);
  
	low_level_init(netif);
	etharp_init();
	/* eth_send_packet(junk,sizeof(junk)); */
	return ERR_OK;
}

int easymacif_select(struct netif *netif)
{
	if (eth_get_status() & EEMAC_RECV_DONE) {	       
		easymacif_input(netif);
		return 0;
	}
	return -1;
}
